from dataclasses import dataclass

from memray import AllocatorType
from memray._memray import HighWaterMarkAggregatorTestHarness
from memray._memray import Interval

CALLOC = AllocatorType.CALLOC
MALLOC = AllocatorType.MALLOC
VALLOC = AllocatorType.VALLOC
FREE = AllocatorType.FREE
MMAP = AllocatorType.MMAP
MUNMAP = AllocatorType.MUNMAP


@dataclass(frozen=True)
class Contribution:
    n_allocations_in_high_water_mark: int
    bytes_in_high_water_mark: int
    n_allocations_leaked: int
    bytes_leaked: int


@dataclass(frozen=True)
class Location:
    tid: int
    native_frame_id: int
    frame_index: int
    native_segment_generation: int


def contribution_by_location_and_allocator(allocations):
    ret = {}
    for alloc in allocations:
        loc = Location(**{k: alloc[k] for k in Location.__dataclass_fields__.keys()})
        contribution = Contribution(
            **{k: alloc[k] for k in Contribution.__dataclass_fields__.keys()}
        )
        key = loc, alloc["allocator"]
        assert key not in ret
        ret[key] = contribution
    return ret


def test_no_allocations_at_start():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()

    # WHEN
    # THEN
    assert 0 == tester.get_current_heap_size()
    assert [] == tester.get_allocations()
    assert [0] == list(tester.high_water_mark_bytes_by_snapshot())
    assert [] == list(tester.get_temporal_allocations())


def test_one_allocation_is_both_high_water_mark_and_leaked():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=4096, size=1234)

    # THEN
    assert 1234 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, allocator)] == Contribution(1, 1234, 1, 1234)
    assert len(contributions) == 1


def test_one_freed_allocation_is_high_water_mark_but_not_leaked():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.FREE, address=4096, size=0
    )

    # THEN
    assert 0 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, allocator)] == Contribution(1, 1234, 0, 0)
    assert len(contributions) == 1


def test_zero_byte_allocation():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=8192, size=0)

    # THEN
    assert 1234 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, allocator)] == Contribution(2, 1234, 2, 1234)
    assert len(contributions) == 1


def test_freeing_one_of_two_high_water_mark_allocations_at_the_same_location():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=8192, size=4321)
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.FREE, address=4096, size=0
    )

    # THEN
    assert 4321 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, allocator)] == Contribution(2, 1234 + 4321, 1, 4321)
    assert len(contributions) == 1


def test_freeing_one_of_two_high_water_mark_allocations_at_different_locations():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=2,
        native_frame_id=8,
        frame_index=10,
        native_segment_generation=12,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(**loc2.__dict__, allocator=allocator, address=8192, size=4321)
    tester.add_allocation(
        **loc1.__dict__, allocator=AllocatorType.FREE, address=4096, size=0
    )

    # THEN
    assert 4321 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc1, allocator)] == Contribution(1, 1234, 0, 0)
    assert contributions[(loc2, allocator)] == Contribution(1, 4321, 1, 4321)
    assert len(contributions) == 2


def test_allocation_freed_before_high_water_mark():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.FREE, address=4096, size=0
    )
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=8192, size=4321)

    # THEN
    assert 4321 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, allocator)] == Contribution(1, 4321, 1, 4321)
    assert len(contributions) == 1


def test_allocation_made_and_leaked_after_high_water_mark():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=2,
        native_frame_id=8,
        frame_index=10,
        native_segment_generation=12,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=8192, size=4321)
    tester.add_allocation(
        **loc1.__dict__, allocator=AllocatorType.FREE, address=8192, size=0
    )
    tester.add_allocation(**loc2.__dict__, allocator=allocator, address=4096, size=1234)

    # THEN
    assert 1234 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc1, allocator)] == Contribution(1, 4321, 0, 0)
    assert contributions[(loc2, allocator)] == Contribution(0, 0, 1, 1234)
    assert len(contributions) == 2


def test_allocation_made_and_freed_after_high_water_mark():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=2,
        native_frame_id=8,
        frame_index=10,
        native_segment_generation=12,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=8192, size=4321)
    tester.add_allocation(
        **loc1.__dict__, allocator=AllocatorType.FREE, address=8192, size=0
    )
    tester.add_allocation(**loc2.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(
        **loc2.__dict__, allocator=AllocatorType.FREE, address=4096, size=0
    )

    # THEN
    assert 0 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc1, allocator)] == Contribution(1, 4321, 0, 0)
    assert len(contributions) == 1


def test_allocation_made_and_freed_between_high_water_marks():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=2,
        native_frame_id=8,
        frame_index=10,
        native_segment_generation=12,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=8192, size=4321)
    tester.add_allocation(
        **loc1.__dict__, allocator=AllocatorType.FREE, address=8192, size=0
    )
    tester.add_allocation(**loc2.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(
        **loc2.__dict__, allocator=AllocatorType.FREE, address=4096, size=0
    )
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=8192, size=4321)
    tester.add_allocation(
        **loc1.__dict__, allocator=AllocatorType.FREE, address=8192, size=0
    )

    # THEN
    assert 0 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc1, allocator)] == Contribution(1, 4321, 0, 0)
    assert len(contributions) == 1


def test_allocation_made_between_high_water_marks_and_freed_after_high_water_mark():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=2,
        native_frame_id=8,
        frame_index=10,
        native_segment_generation=12,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=8192, size=4321)
    tester.add_allocation(
        **loc1.__dict__, allocator=AllocatorType.FREE, address=8192, size=0
    )
    tester.add_allocation(**loc2.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=8192, size=4321)
    tester.add_allocation(
        **loc1.__dict__, allocator=AllocatorType.FREE, address=8192, size=0
    )
    tester.add_allocation(
        **loc2.__dict__, allocator=AllocatorType.FREE, address=4096, size=0
    )

    # THEN
    assert 0 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc1, allocator)] == Contribution(1, 4321, 0, 0)
    assert contributions[(loc2, allocator)] == Contribution(1, 1234, 0, 0)
    assert len(contributions) == 2


def test_allocation_made_between_high_water_marks_and_leaked():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=2,
        native_frame_id=8,
        frame_index=10,
        native_segment_generation=12,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=8192, size=4321)
    tester.add_allocation(
        **loc1.__dict__, allocator=AllocatorType.FREE, address=8192, size=0
    )
    tester.add_allocation(**loc2.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=8192, size=4321)

    # THEN
    assert 1234 + 4321 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc1, allocator)] == Contribution(1, 4321, 1, 4321)
    assert contributions[(loc2, allocator)] == Contribution(1, 1234, 1, 1234)
    assert len(contributions) == 2


def test_different_allocators_at_one_location():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.MALLOC, address=4096, size=1234
    )
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.VALLOC, address=8192, size=4321
    )

    # THEN
    assert 1234 + 4321 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, AllocatorType.MALLOC)] == Contribution(1, 1234, 1, 1234)
    assert contributions[(loc, AllocatorType.VALLOC)] == Contribution(1, 4321, 1, 4321)
    assert len(contributions) == 2


def test_same_stack_in_different_threads():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=40,
        frame_index=50,
        native_segment_generation=60,
    )
    loc2 = Location(
        tid=2,
        native_frame_id=40,
        frame_index=50,
        native_segment_generation=60,
    )

    # WHEN
    tester.add_allocation(
        **loc1.__dict__, allocator=AllocatorType.MALLOC, address=4096, size=1234
    )
    tester.add_allocation(
        **loc2.__dict__, allocator=AllocatorType.MALLOC, address=8192, size=2468
    )

    # THEN
    assert 1234 + 2468 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc1, AllocatorType.MALLOC)] == Contribution(1, 1234, 1, 1234)
    assert contributions[(loc2, AllocatorType.MALLOC)] == Contribution(1, 2468, 1, 2468)
    assert len(contributions) == 2


def test_completely_freed_range():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    allocator = AllocatorType.MMAP

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.MUNMAP, address=4096, size=1234
    )

    # THEN
    assert 0 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, allocator)] == Contribution(1, 1234, 0, 0)
    assert len(contributions) == 1


def test_shrunk_range():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    allocator = AllocatorType.MMAP

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.MUNMAP, address=4096, size=1000
    )

    # THEN
    assert 234 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, allocator)] == Contribution(1, 1234, 1, 234)
    assert len(contributions) == 1


def test_shrunk_then_freed_range():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    allocator = AllocatorType.MMAP

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.MUNMAP, address=4096 + 1000, size=234
    )
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.MUNMAP, address=4096, size=1000
    )

    # THEN
    assert 0 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, allocator)] == Contribution(1, 1234, 0, 0)
    assert len(contributions) == 1


def test_split_range():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    allocator = AllocatorType.MMAP

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.MUNMAP, address=4096 + 100, size=100
    )

    # THEN
    assert 1234 - 100 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, allocator)] == Contribution(1, 1234, 2, 1134)
    assert len(contributions) == 1


def test_split_then_freed_range():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    allocator = AllocatorType.MMAP

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=allocator, address=4096, size=1234)
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.MUNMAP, address=4096 + 100, size=100
    )
    tester.add_allocation(
        **loc.__dict__, allocator=AllocatorType.MUNMAP, address=4096, size=1234
    )

    # THEN
    assert 0 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc, allocator)] == Contribution(1, 1234, 0, 0)
    assert len(contributions) == 1


def test_reporting_on_true_high_water_mark_that_was_in_a_past_snapshot():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=2,
        native_frame_id=8,
        frame_index=10,
        native_segment_generation=12,
    )
    allocator = AllocatorType.CALLOC

    # WHEN
    tester.capture_snapshot()
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=8192, size=4321)
    tester.capture_snapshot()
    tester.add_allocation(**loc2.__dict__, allocator=allocator, address=4096, size=100)
    tester.capture_snapshot()
    tester.add_allocation(
        **loc1.__dict__, allocator=AllocatorType.FREE, address=8192, size=0
    )
    tester.add_allocation(
        **loc2.__dict__, allocator=AllocatorType.FREE, address=4096, size=0
    )
    tester.capture_snapshot()
    tester.add_allocation(**loc2.__dict__, allocator=allocator, address=4096, size=2000)
    tester.add_allocation(
        **loc2.__dict__, allocator=AllocatorType.FREE, address=4096, size=0
    )
    tester.add_allocation(**loc1.__dict__, allocator=allocator, address=8192, size=1000)

    # THEN
    assert 1000 == tester.get_current_heap_size()
    contributions = contribution_by_location_and_allocator(tester.get_allocations())
    assert contributions[(loc1, allocator)] == Contribution(1, 4321, 1, 1000)
    assert contributions[(loc2, allocator)] == Contribution(1, 100, 0, 0)
    assert len(contributions) == 2


def test_one_allocation_before_first_snapshot():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)

    # THEN
    (alloc,) = tester.get_temporal_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.stack_id == 5
    assert alloc.tid == 1
    assert alloc.intervals == [
        Interval(0, None, 1, 1234),
    ]


def test_one_allocation_after_first_snapshot():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)

    # THEN
    (alloc,) = tester.get_temporal_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.stack_id == 5
    assert alloc.tid == 1
    assert alloc.intervals == [
        Interval(1, None, 1, 1234),
    ]


def test_one_allocation_freed_at_high_water_mark_in_second_snapshot():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)

    # THEN
    (alloc,) = tester.get_temporal_allocations()
    assert alloc.allocator == CALLOC
    assert alloc.stack_id == 5
    assert alloc.tid == 1
    assert alloc.intervals == [
        Interval(0, 2, 1, 1234),
        Interval(2, None, 0, 0),
    ]


def test_two_allocations_in_different_snapshots():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=1,
        native_frame_id=7,
        frame_index=8,
        native_segment_generation=9,
    )

    # WHEN
    tester.add_allocation(**loc1.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.capture_snapshot()
    tester.add_allocation(**loc2.__dict__, allocator=CALLOC, address=8192, size=2468)

    # THEN
    (alloc1, alloc2) = tester.get_temporal_allocations()
    assert alloc1.allocator == CALLOC
    assert alloc1.stack_id == 5
    assert alloc1.tid == 1
    assert alloc1.intervals == [
        Interval(0, None, 1, 1234),
    ]

    assert alloc2.allocator == CALLOC
    assert alloc2.stack_id == 8
    assert alloc2.tid == 1
    assert alloc2.intervals == [
        Interval(1, None, 1, 2468),
    ]


def test_one_allocation_freed_before_high_water_mark_in_second_snapshot():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc1 = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )
    loc2 = Location(
        tid=1,
        native_frame_id=7,
        frame_index=8,
        native_segment_generation=9,
    )

    # WHEN
    tester.add_allocation(**loc1.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.capture_snapshot()
    tester.add_allocation(**loc1.__dict__, allocator=FREE, address=4096, size=0)
    tester.add_allocation(**loc2.__dict__, allocator=CALLOC, address=4096, size=2468)

    # THEN
    (alloc1, alloc2) = tester.get_temporal_allocations()
    assert alloc1.allocator == CALLOC
    assert alloc1.stack_id == 5
    assert alloc1.tid == 1
    assert alloc1.intervals == [
        Interval(0, 1, 1, 1234),
        Interval(1, None, 0, 0),
    ]

    assert alloc2.allocator == CALLOC
    assert alloc2.stack_id == 8
    assert alloc2.tid == 1
    assert alloc2.intervals == [
        Interval(1, None, 1, 2468),
    ]


def test_allocations_freed_over_two_snapshots():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=2468)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)
    tester.capture_snapshot()

    # THEN
    (alloc,) = tester.get_temporal_allocations()

    assert alloc.allocator == CALLOC
    assert alloc.stack_id == 5
    assert alloc.tid == 1
    assert alloc.intervals == [
        Interval(0, 2, 2, 1234 + 2468),
        Interval(2, 3, 1, 2468),
        Interval(3, None, 0, 0),
    ]


def test_allocations_freed_over_two_non_adjacent_snapshots():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=2468)
    tester.capture_snapshot()
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)
    tester.capture_snapshot()
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)
    tester.capture_snapshot()

    # THEN
    (alloc,) = tester.get_temporal_allocations()

    assert alloc.allocator == CALLOC
    assert alloc.stack_id == 5
    assert alloc.tid == 1
    assert alloc.intervals == [
        Interval(0, 3, 2, 1234 + 2468),
        Interval(3, 5, 1, 2468),
        Interval(5, None, 0, 0),
    ]


def test_allocation_after_high_water_mark_in_current_snapshot():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=2468)
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=1234)

    # THEN
    (alloc,) = tester.get_temporal_allocations()

    assert alloc.allocator == CALLOC
    assert alloc.stack_id == 5
    assert alloc.tid == 1
    assert alloc.intervals == [
        Interval(0, 1, 1, 2468),
        Interval(1, None, 1, 1234),
    ]


def test_allocation_after_high_water_mark_in_historical_snapshot():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=2468)
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=1234)
    tester.capture_snapshot()

    # THEN
    (alloc,) = tester.get_temporal_allocations()

    assert alloc.allocator == CALLOC
    assert alloc.stack_id == 5
    assert alloc.tid == 1
    assert alloc.intervals == [
        Interval(0, 1, 1, 2468),
        Interval(1, None, 1, 1234),
    ]


def test_allocation_and_deallocation_after_high_water_mark():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=2468)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1230)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=2460)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=0, size=1)

    # THEN
    (alloc,) = tester.get_temporal_allocations()

    assert alloc.allocator == CALLOC
    assert alloc.stack_id == 5
    assert alloc.tid == 1
    assert alloc.intervals == [
        Interval(0, 1, 2, 2468 + 1234),
        Interval(1, None, 3, 1230 + 2460 + 1),
    ]


def test_allocation_and_deallocation_across_multiple_snapshots():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=2468)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=4096, size=0)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1230)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=2460)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=0, size=1)

    # THEN
    (alloc,) = tester.get_temporal_allocations()

    assert alloc.allocator == CALLOC
    assert alloc.stack_id == 5
    assert alloc.tid == 1
    assert alloc.intervals == [
        Interval(0, 1, 2, 2468 + 1234),
        Interval(1, 2, 1, 1234),
        Interval(2, None, 3, 1230 + 2460 + 1),
    ]


def test_allocation_and_deallocation_across_multiple_snapshots_with_other_allocators():
    # GIVEN
    tester = HighWaterMarkAggregatorTestHarness()
    loc = Location(
        tid=1,
        native_frame_id=4,
        frame_index=5,
        native_segment_generation=6,
    )

    # WHEN
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=8192, size=2468)
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=4096, size=1234)
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=MALLOC, address=8192, size=1230)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=VALLOC, address=0, size=1)
    tester.capture_snapshot()
    tester.add_allocation(**loc.__dict__, allocator=CALLOC, address=1000, size=42)
    tester.add_allocation(**loc.__dict__, allocator=FREE, address=8192, size=0)

    # THEN
    (alloc1, alloc2, alloc3) = tester.get_temporal_allocations()

    assert alloc1.allocator == MALLOC
    assert alloc1.stack_id == 5
    assert alloc1.tid == 1
    assert alloc1.intervals == [
        Interval(1, 4, 1, 1230),
        Interval(4, None, 0, 0),
    ]

    assert alloc2.allocator == CALLOC
    assert alloc2.stack_id == 5
    assert alloc2.tid == 1
    assert alloc2.intervals == [
        Interval(0, 1, 2, 2468 + 1234),
        Interval(1, 3, 1, 1234),
        Interval(3, None, 2, 1234 + 42),
    ]

    assert alloc3.allocator == VALLOC
    assert alloc3.stack_id == 5
    assert alloc3.tid == 1
    assert alloc3.intervals == [
        Interval(2, None, 1, 1),
    ]



from pathlib import Path

import pytest

from memray.reporters.module_tools import ModuleResolver
from memray.reporters.module_tools import _path_to_module
from memray.reporters.module_tools import get_python_path_info


@pytest.fixture(scope="module")
def path_info():
    return get_python_path_info()


@pytest.fixture
def resolver(path_info):
    return ModuleResolver(path_info)


def test_path_info_has_expected_keys():
    info = get_python_path_info()
    assert "stdlib" in info
    assert "site_packages" in info
    assert "sys_path" in info


def test_path_info_stdlib_points_to_sysconfig_libdest():
    import sysconfig

    info = get_python_path_info()
    assert info["stdlib"] is not None
    assert info["stdlib"] == Path(sysconfig.get_config_var("LIBDEST"))


def test_path_info_site_packages_is_list_of_paths():
    info = get_python_path_info()
    assert isinstance(info["site_packages"], list)
    for p in info["site_packages"]:
        assert isinstance(p, Path)


def test_path_info_sys_path_is_list_of_paths():
    info = get_python_path_info()
    assert isinstance(info["sys_path"], list)
    for p in info["sys_path"]:
        assert isinstance(p, Path)


def test_path_info_includes_user_site_packages(tmp_path):
    import unittest.mock

    user_site = tmp_path / "user-site-packages"
    user_site.mkdir()
    with unittest.mock.patch("site.getusersitepackages", return_value=str(user_site)):
        info = get_python_path_info()
    assert Path(user_site) in info["site_packages"]


@pytest.mark.parametrize(
    "path, expected",
    [
        ("mymodule.py", "mymodule"),
        ("pandas/io/parsers.py", "pandas.io.parsers"),
        ("requests/__init__.py", "requests"),
        ("pkg/utils", "pkg.utils"),
        ("numpy/core/__init__.py", "numpy.core"),
    ],
)
def test_path_to_module(path, expected):
    assert _path_to_module(Path(path)) == expected


def test_extract_module_name_and_type_empty(resolver):
    name, mtype = resolver.extract_module_name_and_type("")
    assert name == "unknown"
    assert mtype == "unknown"


def test_extract_module_name_and_type_frozen(resolver):
    name, mtype = resolver.extract_module_name_and_type("<frozen importlib._bootstrap>")
    assert name == "importlib._bootstrap"
    assert mtype == "stdlib"


def test_extract_module_name_and_type_stdlib(resolver, path_info):
    stdlib_file = str(path_info["stdlib"] / "json" / "__init__.py")
    name, mtype = resolver.extract_module_name_and_type(stdlib_file)
    assert mtype == "stdlib"
    assert "json" in name


def test_extract_module_name_and_type_site_packages(resolver):
    name, mtype = resolver.extract_module_name_and_type(pytest.__file__)
    assert mtype == "site-packages"
    assert "pytest" in name


def test_extract_module_name_and_type_project(resolver):
    _, mtype = resolver.extract_module_name_and_type(__file__)
    assert mtype in ("project", "unknown")


def test_extract_module_name_and_type_unknown(resolver):
    name, mtype = resolver.extract_module_name_and_type(
        "/nonexistent/random/mymodule.py"
    )
    assert mtype == "unknown"
    assert name == "mymodule"


def test_extract_module_name_and_type_is_cached(resolver):
    # GIVEN / WHEN
    first = resolver.extract_module_name_and_type(pytest.__file__)
    second = resolver.extract_module_name_and_type(pytest.__file__)

    # THEN
    assert first == second
    assert resolver._cache[pytest.__file__] == first


def test_skips_stdlib_leaf_frame(resolver, path_info):
    # GIVEN
    stdlib_path = str(path_info["stdlib"] / "json" / "__init__.py")
    stack = [
        ("loads", stdlib_path, 346),
        ("safe_load", pytest.__file__, 10),
    ]

    # THEN
    assert resolver.get_module_for_stack(stack) == "pytest"


def test_skips_frozen_stdlib_frame(resolver):
    # GIVEN
    stack = [
        ("_find", "<frozen importlib._bootstrap>", 100),
        ("safe_load", pytest.__file__, 10),
    ]

    # THEN
    assert resolver.get_module_for_stack(stack) == "pytest"


def test_returns_non_stdlib_leaf_immediately(resolver, path_info):
    # GIVEN
    stdlib_path = str(path_info["stdlib"] / "json" / "__init__.py")
    stack = [
        ("safe_load", pytest.__file__, 10),
        ("loads", stdlib_path, 346),
    ]

    # THEN
    assert resolver.get_module_for_stack(stack) == "pytest"


def test_all_stdlib_returns_root(resolver, path_info):
    # GIVEN
    stdlib = path_info["stdlib"]
    stack = [
        ("loads", str(stdlib / "json" / "__init__.py"), 346),
        ("decode", str(stdlib / "json" / "decoder.py"), 337),
    ]

    # THEN
    assert resolver.get_module_for_stack(stack) == "<root>"


def test_empty_stack_returns_root(resolver):
    assert resolver.get_module_for_stack([]) == "<root>"


def _memray_frame_path(path_info):
    site_packages = path_info["site_packages"]
    if not site_packages:
        pytest.skip("No site-packages directory to anchor a memray path")
    return str(site_packages[0] / "memray" / "commands" / "run.py")


def test_memray_own_frames_return_root(resolver, path_info):
    # GIVEN
    stdlib_path = str(path_info["stdlib"] / "json" / "__init__.py")
    stack = [
        ("loads", stdlib_path, 346),
        ("_run_tracker", _memray_frame_path(path_info), 100),
    ]

    # THEN
    # memray's own driver frames must not be attributed as a module; they are
    # treated as the root boundary.
    assert resolver.get_module_for_stack(stack) == "<root>"


def test_memray_frames_do_not_shadow_user_frames(resolver, path_info):
    # GIVEN
    stack = [
        ("work", pytest.__file__, 10),
        ("_run_tracker", _memray_frame_path(path_info), 100),
    ]

    # THEN
    assert resolver.get_module_for_stack(stack) == "pytest"



import os
import threading

import pytest

from memray import FileDestination
from memray import FileReader
from memray import Tracker
from memray._test import MemoryAllocator


def test_rejects_different_header_magic(tmp_path):
    # GIVEN
    output = tmp_path / "test.bin"
    allocator = MemoryAllocator()

    # WHEN
    # Create a valid allocation record file
    with Tracker(output):
        allocator.valloc(1024)

    # Change the header magic (same length as "memray")
    with output.open("rb+") as f:
        f.write(b"badxxx")

    # THEN
    with pytest.raises(
        OSError, match="does not look like a binary generated by memray"
    ):
        FileReader(output).get_allocation_records()


def test_rejects_different_header_version(tmp_path):
    # GIVEN
    output = tmp_path / "test.bin"
    allocator = MemoryAllocator()

    # WHEN
    # Create a valid allocation record file
    destination = FileDestination(output, overwrite=False, compress_on_exit=False)
    with Tracker(destination=destination):
        allocator.valloc(1024)

    # Change the header version to zero
    with output.open("rb+") as f:
        f.seek(7)
        f.write(b"\0")

    # THEN
    with pytest.raises(OSError, match="incompatible with this version"):
        FileReader(output).get_allocation_records()


def test_filereader_fails_to_open_file(tmp_path):
    """This checks that we throw in the FileSource C++ ctor when we fail to open the stream."""
    # GIVEN
    test_file = tmp_path / "test.bin"
    test_file.touch(mode=000)

    try:
        test_file.read_text()
    except OSError:
        pass
    else:  # pragma: no cover
        pytest.skip("The current user can ignore file permissions")

    # WHEN/THEN
    with pytest.raises(OSError, match="Could not open file"):
        FileReader(test_file)


def test_read_pid(tmp_path):
    # GIVEN
    output = tmp_path / "test.bin"
    allocator = MemoryAllocator()

    # WHEN
    with Tracker(output):
        allocator.valloc(1024)

    # THEN
    assert FileReader(output).metadata.pid == os.getpid()


def test_read_tid(tmp_path):
    # GIVEN
    output = tmp_path / "test.bin"
    allocator = MemoryAllocator()

    def func():
        allocator.valloc(1024)

    # WHEN
    t = threading.Thread(target=func)
    with Tracker(output):
        func()
        t.start()
        t.join()

    # THEN
    reader = FileReader(output)
    main_tid = reader.metadata.main_thread_id
    all_allocations = reader.get_allocation_records()
    all_tids = tuple(allocation.tid for allocation in all_allocations)
    assert main_tid in set(all_tids)
    # The main thread should be the first one in the list
    assert main_tid == all_tids[0]



"""Test the performance of pprint.PrettyPrinter.

This benchmark was available as `python -m pprint` until Python 3.12.

Authors: Fred Drake (original), Oleg Iarygin (pyperformance port).
"""

import pyperf
from pprint import PrettyPrinter
from memray_helper import get_tracker


printable = [("string", (1, 2), [3, 4], {5: 6, 7: 8})] * 100_000
p = PrettyPrinter()


def format(*args, **kwrags):
    with get_tracker():
        p.pformat(*args, **kwrags)


def safe_repr(*args, **kwargs):
    with get_tracker():
        p._safe_repr(*args, **kwargs)


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata["description"] = "pprint benchmark"

    if hasattr(p, "_safe_repr"):
        runner.bench_func("pprint_safe_repr", safe_repr, printable, {}, None, 0)
    runner.bench_func("pprint_pformat", format, printable)



import mmap
import os
import sys
import threading
from pathlib import Path

import pytest

from memray import AllocatorType
from memray import FileReader
from memray import Tracker
from memray._test import MemoryAllocator
from memray._test import _cython_nested_allocation
from memray._test import allocate_without_gil_held
from memray._test import function_caller


def alloc_func3(allocator):
    x = 1
    allocator.valloc(123456)
    x = 2
    allocator.free()
    x = 3
    return x


def alloc_func2(allocator):
    y = 1
    alloc_func3(allocator)
    y = 2
    return y


def alloc_func1(allocator):
    z = 1
    alloc_func2(allocator)
    z = 2
    return z


def test_traceback(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        alloc_func1(allocator)
    records = list(FileReader(output).get_allocation_records())

    # THEN

    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]
    assert len(allocs) == 1
    (alloc,) = allocs
    traceback = list(alloc.stack_trace())
    assert traceback[-4:] == [
        ("alloc_func3", __file__, 20),
        ("alloc_func2", __file__, 29),
        ("alloc_func1", __file__, 36),
        ("test_traceback", __file__, 49),
    ]
    frees = [
        record
        for record in records
        if record.allocator == AllocatorType.FREE and record.address == alloc.address
    ]
    assert len(frees) == 1
    (free,) = frees
    with pytest.raises(NotImplementedError):
        free.stack_trace()


def test_traceback_for_high_watermark(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        alloc_func1(allocator)
    records = list(FileReader(output).get_high_watermark_allocation_records())

    # THEN

    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]
    assert len(allocs) == 1
    (alloc,) = allocs
    traceback = list(alloc.stack_trace())
    assert traceback[-4:] == [
        ("alloc_func3", __file__, 20),
        ("alloc_func2", __file__, 29),
        ("alloc_func1", __file__, 36),
        ("test_traceback_for_high_watermark", __file__, 83),
    ]


def test_traceback_iteration_does_not_depend_on_the_order_of_elements(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        alloc_func1(allocator)
        alloc_func1(allocator)

    # THEN

    records = list(FileReader(output).get_allocation_records())
    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]
    alloc1, alloc2 = allocs
    traceback1 = list(alloc1.stack_trace())
    traceback2 = list(alloc2.stack_trace())

    records = list(FileReader(output).get_allocation_records())
    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]
    alloc1, alloc2 = allocs
    assert traceback2 == list(alloc2.stack_trace())
    assert traceback1 == list(alloc1.stack_trace())


def test_cython_traceback(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        _cython_nested_allocation(allocator.valloc, 1234)
    allocator.free()
    records = list(FileReader(output).get_allocation_records())

    # THEN

    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]
    assert len(allocs) == 2
    alloc1, alloc2 = allocs

    traceback = list(alloc1.stack_trace())
    assert traceback == [
        ("valloc", sys.modules["memray._test"].__file__, 50),
        ("test_cython_traceback", __file__, 134),
    ]

    traceback = list(alloc2.stack_trace())
    assert traceback == [
        ("test_cython_traceback", __file__, 134),
    ]

    frees = [
        record
        for record in records
        if record.allocator == AllocatorType.FREE and record.address == alloc2.address
    ]
    assert len(frees) == 1
    (free,) = frees
    with pytest.raises(NotImplementedError):
        free.stack_trace()


def test_large_number_of_frame_pops_between_subsequent_allocations(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"

    def allocate_deep(depth):
        if depth <= 1:
            return mmap.mmap(-1, 1234)
        return allocate_deep(depth - 1)

    # WHEN
    # Note: we don't actually care about the native stacks, but we use
    # native_traces=True to ensure that the allocation we care about inside
    # of `mmap.mmap` has a different stack than any allocation that the
    # interpreter itself performs. Otherwise, our high water mark aggregator
    # could combine the mmap we care about with other allocations performed
    # inside the interpreter that happen to share the same Python stack.
    with Tracker(output, native_traces=True):
        with allocate_deep(20):
            with mmap.mmap(-1, 12345):
                pass
    records = list(FileReader(output).get_high_watermark_allocation_records())

    # THEN
    allocs = [
        record
        for record in records
        if record.allocator == AllocatorType.MMAP and record.size == 1234
    ]
    assert len(allocs) == 1
    (alloc,) = allocs
    traceback = list(alloc.stack_trace())
    assert len(traceback) == 21

    allocs = [
        record
        for record in records
        if record.allocator == AllocatorType.MMAP and record.size == 12345
    ]
    assert len(allocs) == 1
    (alloc,) = allocs
    traceback = list(alloc.stack_trace())
    assert len(traceback) == 1


def test_records_can_be_retrieved_twice(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        alloc_func1(allocator)

    # THEN

    reader = FileReader(output)
    records1 = list(reader.get_allocation_records())
    records2 = list(reader.get_allocation_records())

    assert records1 == records2


def test_high_watermark_records_can_be_retrieved_twice(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        alloc_func1(allocator)

    # THEN

    reader = FileReader(output)
    records1 = list(reader.get_high_watermark_allocation_records())
    records2 = list(reader.get_high_watermark_allocation_records())

    assert records1 == records2


def test_traceback_can_be_retrieved_twice(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        alloc_func1(allocator)

    # THEN

    records = list(FileReader(output).get_allocation_records())
    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]
    (alloc,) = allocs
    traceback1 = list(alloc.stack_trace())
    traceback2 = list(alloc.stack_trace())
    assert traceback1 == traceback2


def test_traceback_for_high_watermark_records_can_be_retrieved_twice(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        alloc_func1(allocator)

    # THEN

    reader = FileReader(output)
    records = list(reader.get_high_watermark_allocation_records())
    (alloc,) = records
    traceback1 = list(alloc.stack_trace())
    records = list(reader.get_high_watermark_allocation_records())
    (alloc,) = records
    traceback2 = list(alloc.stack_trace())

    assert traceback1 == traceback2


def test_profile_function_is_restored_after_tracking(tmpdir):
    # GIVEN
    def profilefunc(*args):
        pass

    output = Path(tmpdir) / "test.bin"

    # WHEN

    sys.setprofile(profilefunc)

    with Tracker(output):
        assert sys.getprofile() != profilefunc

    # THEN
    assert sys.getprofile() == profilefunc


def test_initial_tracking_frames_are_correctly_populated(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    def foo():
        allocator.valloc(1234)
        allocator.free()

    # WHEN

    with Tracker(output):
        foo()
    records = list(FileReader(output).get_allocation_records())

    # THEN

    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]
    assert len(allocs) == 1
    (alloc,) = allocs
    traceback = [frame[0] for frame in alloc.stack_trace()]
    assert traceback[-4:] == [
        "valloc",
        "foo",
        "test_initial_tracking_frames_are_correctly_populated",
    ]


def test_restart_tracing_function_gets_correctly_the_frames(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    def foo():
        allocator.valloc(1234)
        allocator.free()

    # WHEN

    # Do some prelininary tracing to populate the initial frames
    with Tracker(output):
        foo()

    output.unlink()

    def bar():
        foo()

    # Do another *independent* round of tracking. The previous frames
    # should not interfere with this tracing.
    with Tracker(output):
        bar()
    records = list(FileReader(output).get_allocation_records())

    # THEN

    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]
    assert len(allocs) == 1
    (alloc,) = allocs
    traceback = [frame[0] for frame in alloc.stack_trace()]
    assert traceback[-5:] == [
        "valloc",
        "foo",
        "bar",
        "test_restart_tracing_function_gets_correctly_the_frames",
    ]


def test_num_records(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        alloc_func1(allocator)
        alloc_func1(allocator)

    # THEN
    reader = FileReader(output)
    n_records = len(list(reader.get_allocation_records()))
    assert n_records == reader.metadata.total_allocations


def test_allocations_in_root_frame_have_correct_line_number(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    first = second = None

    # WHEN
    with Tracker(output):
        first = mmap.mmap(-1, 1)
        second = mmap.mmap(-1, 1)
        del first
        del second

    # THEN
    records = list(FileReader(output).get_allocation_records())
    print(records)
    allocs = [
        record
        for record in records
        if record.allocator == AllocatorType.MMAP and record.size == 1
    ]

    assert len(allocs) == 2
    alloc1, alloc2 = allocs
    func1, file1, line1 = alloc1.stack_trace()[0]
    func2, file2, line2 = alloc2.stack_trace()[0]
    assert func1 == "test_allocations_in_root_frame_have_correct_line_number"
    assert func2 == "test_allocations_in_root_frame_have_correct_line_number"
    assert file1 == __file__
    assert file2 == __file__
    assert abs(line1 - line2) == 1


def test_equal_stack_traces_compare_equal(tmpdir):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        for _ in range(2):
            alloc_func1(allocator)
            alloc_func2(allocator)

    # THEN

    records = list(FileReader(output).get_allocation_records())
    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]

    assert len(allocs) == 4
    first_alloc1, first_alloc2, second_alloc1, second_alloc2 = allocs

    assert first_alloc1.stack_id == second_alloc1.stack_id
    assert first_alloc1.stack_trace() == second_alloc1.stack_trace()
    assert first_alloc2.stack_id == second_alloc2.stack_id
    assert first_alloc2.stack_trace() == second_alloc2.stack_trace()

    assert first_alloc1.stack_id != first_alloc2.stack_id
    assert second_alloc1.stack_id != second_alloc2.stack_id


def test_identical_stack_traces_started_in_different_lines_in_the_root_do_not_compare_equal(
    tmpdir,
):  # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    with Tracker(output):
        alloc_func1(allocator)
        alloc_func2(allocator)
        alloc_func1(allocator)
        alloc_func2(allocator)

    # THEN

    records = list(FileReader(output).get_allocation_records())
    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]

    assert len(allocs) == 4
    first_alloc1, first_alloc2, second_alloc1, second_alloc2 = allocs

    assert first_alloc1.stack_id != second_alloc1.stack_id
    assert first_alloc1.stack_trace()[:-1] == second_alloc1.stack_trace()[:-1]
    assert first_alloc1.stack_trace()[-1] != second_alloc1.stack_trace()[-1]
    assert first_alloc2.stack_id != second_alloc2.stack_id
    assert first_alloc2.stack_trace()[:-1] == second_alloc2.stack_trace()[:-1]
    assert first_alloc2.stack_trace()[-1] != second_alloc2.stack_trace()[-1]

    assert first_alloc1.stack_id != first_alloc2.stack_id
    assert second_alloc1.stack_id != second_alloc2.stack_id


def test_identical_stack_traces_started_in_different_lines_in_a_function_do_not_compare_equal(
    tmpdir,
):
    # GIVEN
    allocator = MemoryAllocator()
    output = Path(tmpdir) / "test.bin"

    # WHEN

    def foo():
        alloc_func1(allocator)
        alloc_func2(allocator)
        alloc_func1(allocator)
        alloc_func2(allocator)

    with Tracker(output):
        foo()

    # THEN

    records = list(FileReader(output).get_allocation_records())
    allocs = [record for record in records if record.allocator == AllocatorType.VALLOC]

    assert len(allocs) == 4
    first_alloc1, first_alloc2, second_alloc1, second_alloc2 = allocs

    assert first_alloc1.stack_id != second_alloc1.stack_id
    assert first_alloc1.stack_trace() != second_alloc1.stack_trace()
    assert first_alloc2.stack_id != second_alloc2.stack_id
    assert first_alloc2.stack_trace() != second_alloc2.stack_trace()

    assert first_alloc1.stack_id != first_alloc2.stack_id
    assert second_alloc1.stack_id != second_alloc2.stack_id


def test_allocation_in_thread_started_before_tracking_starts(tmp_path):
    """Test capturing the stack of a thread started before tracking started.

    The intended execution flow is:
    Main Thread          Background Thread
    -----------          -----------------
    Start thread
                         Call thread_body
    Install tracker
                         Call func1
                         Perform an allocation
                         Exit
    Join thread
    Uninstall tracker
    """
    # GIVEN
    thread_body_entered = threading.Event()
    tracker_installed = threading.Event()
    allocator = MemoryAllocator()
    output = tmp_path / "test.bin"

    def thread_body():
        thread_body_entered.set()
        tracker_installed.wait()
        func1()

    def func1():
        allocator.valloc(1234)
        allocator.free()

    # WHEN
    bg_thread = threading.Thread(target=thread_body)
    bg_thread.start()

    thread_body_entered.wait()
    with Tracker(output):
        tracker_installed.set()
        bg_thread.join()

    # THEN
    allocations = list(FileReader(output).get_allocation_records())

    vallocs = [
        event
        for event in allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    assert len(vallocs) == 1

    funcs = [frame[0] for frame in vallocs[0].stack_trace()]
    assert funcs == [
        "valloc",
        "func1",
        "thread_body",
        "run",
        "_bootstrap_inner",
        "_bootstrap",
    ]


def test_allocation_in_thread_before_reacquiring_gil_after_tracking_starts(tmp_path):
    """
    The intended execution flow is:
    Main Thread          Background Thread
    -----------          -----------------
    Start thread
                         Call allocate_without_gil_held
                         release GIL
    Install tracker
                         Perform an allocation for 1234 bytes
                         acquire GIL
                         Perform an allocation for 4321 bytes
                         Exit
    Join thread
    Uninstall tracker
    """
    # GIVEN
    wake_up_main_r, wake_up_main_w = os.pipe()
    wake_up_thread_r, wake_up_thread_w = os.pipe()
    output = tmp_path / "test.bin"

    def thread_body():
        allocate_without_gil_held(wake_up_main_w, wake_up_thread_r)

    # WHEN
    bg_thread = threading.Thread(target=thread_body)
    bg_thread.start()

    os.read(wake_up_main_r, 1)
    with Tracker(output):
        os.write(wake_up_thread_w, b"x")
        bg_thread.join()

    # THEN
    allocations = list(FileReader(output).get_allocation_records())

    vallocs = [
        event for event in allocations if event.allocator == AllocatorType.VALLOC
    ]
    assert len(vallocs) == 2

    funcs1 = [frame[0] for frame in vallocs[0].stack_trace()]
    funcs2 = [frame[0] for frame in vallocs[1].stack_trace()]
    expected = ["thread_body", "run", "_bootstrap_inner", "_bootstrap"]
    assert funcs1 == funcs2 == expected


def test_thread_surviving_multiple_trackers(tmp_path):
    # GIVEN
    orig_tracker_used = threading.Event()
    new_tracker_installed = threading.Event()
    allocator = MemoryAllocator()
    output1 = tmp_path / "test.bin.1"
    output2 = tmp_path / "test.bin.2"

    def deeper_function():
        allocator.valloc(1234)
        allocator.free()
        orig_tracker_used.set()
        new_tracker_installed.wait()
        allocator.valloc(1234)
        allocator.free()

    def tracking_function():
        deeper_function()

    # WHEN
    with Tracker(output1):
        bg_thread = threading.Thread(target=tracking_function)
        bg_thread.start()
        orig_tracker_used.wait()

    with Tracker(output2):
        new_tracker_installed.set()
        bg_thread.join()

    # THEN
    tracker1_allocations = list(FileReader(output1).get_allocation_records())
    tracker2_allocations = list(FileReader(output2).get_allocation_records())

    tracker1_vallocs = [
        event
        for event in tracker1_allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    tracker2_vallocs = [
        event
        for event in tracker2_allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    assert len(tracker1_vallocs) == len(tracker2_vallocs) == 1
    assert tracker1_vallocs[0].stack_trace() != tracker2_vallocs[0].stack_trace()


def test_thread_surviving_multiple_trackers_with_changing_callstack(tmp_path):
    """Test the call stack of a thread changing between two tracking sessions.

    The intended execution flow is:
    Main Thread          Background Thread
    -----------          -----------------
    Install tracker
    Start thread
                         Call thread_body
                         Call func1
                         Perform an allocation
    Uninstall tracker
                         Return from func1
    Install new tracker
                         Call func2
                         Perform an allocation
                         Return from func2
                         Return from thread_body
    Uninstall tracker

    We use a bunch of events to force this order.
    """
    # GIVEN
    allocation_performed_in_func1 = threading.Event()
    tracker_uninstalled = threading.Event()
    returned_from_func1 = threading.Event()
    new_tracker_installed = threading.Event()

    allocator = MemoryAllocator()
    output1 = tmp_path / "test.bin.1"
    output2 = tmp_path / "test.bin.2"

    def thread_body():
        func1()
        returned_from_func1.set()
        new_tracker_installed.wait()
        func2()

    def func1():
        allocator.valloc(1234)
        allocator.free()
        allocation_performed_in_func1.set()
        tracker_uninstalled.wait()

    def func2():
        allocator.valloc(1234)
        allocator.free()

    # WHEN
    with Tracker(output1):
        bg_thread = threading.Thread(target=thread_body)
        bg_thread.start()
        allocation_performed_in_func1.wait()

    tracker_uninstalled.set()
    returned_from_func1.wait()

    with Tracker(output2):
        new_tracker_installed.set()
        bg_thread.join()

    # THEN
    tracker1_allocations = list(FileReader(output1).get_allocation_records())
    tracker2_allocations = list(FileReader(output2).get_allocation_records())

    tracker1_vallocs = [
        event
        for event in tracker1_allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    tracker2_vallocs = [
        event
        for event in tracker2_allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    assert len(tracker1_vallocs) == len(tracker2_vallocs) == 1

    tracker1_funcs = [frame[0] for frame in tracker1_vallocs[0].stack_trace()]
    tracker2_funcs = [frame[0] for frame in tracker2_vallocs[0].stack_trace()]

    common_frames = ["thread_body", "run", "_bootstrap_inner", "_bootstrap"]
    assert tracker1_funcs == ["valloc", "func1"] + common_frames
    assert tracker2_funcs == ["valloc", "func2"] + common_frames


def test_cython_frame_in_pre_existing_thread_stack(tmp_path):
    """Test starting tracking when another thread's stack has Cython frames.

    The intended execution flow is:
    Main Thread          Background Thread
    -----------          -----------------
    Start thread
                         Call thread_body
                         Call function_caller (Cython frame)
                         Call func1
    Install new tracker
                         Perform an allocation
                         Return from func1
                         Return from function_caller
                         Perform an allocation
                         Return from thread_body
    Join thread
    Uninstall tracker
    """
    # GIVEN
    ready_to_install_tracker = threading.Event()
    tracker_installed = threading.Event()

    allocator = MemoryAllocator()
    output = tmp_path / "test.bin"

    def thread_body():
        function_caller(func1)
        allocator.valloc(1234)
        allocator.free()

    def func1():
        ready_to_install_tracker.set()
        tracker_installed.wait()
        allocator.valloc(1234)
        allocator.free()

    # WHEN
    thread = threading.Thread(target=thread_body)
    thread.start()
    ready_to_install_tracker.wait()

    with Tracker(output):
        tracker_installed.set()
        thread.join()

    # THEN
    allocations = list(FileReader(output).get_allocation_records())

    vallocs = [
        event
        for event in allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    assert len(vallocs) == 2

    first, second = vallocs
    alloc1_funcs = [frame[0] for frame in first.stack_trace()]
    alloc2_funcs = [frame[0] for frame in second.stack_trace()]

    # Cython frames called before tracking started aren't in the Python stack.
    assert alloc1_funcs[:3] == ["valloc", "func1", "thread_body"]
    assert alloc2_funcs[:2] == ["valloc", "thread_body"]


def test_cython_frame_in_pre_existing_thread_stack_when_restarting_tracking(tmp_path):
    """Test restarting tracking when another thread's stack has Cython frames.

    The intended execution flow is:
    Main Thread          Background Thread
    -----------          -----------------
    Install tracker
    Start thread
                         Call thread_body
                         Call function_caller (Cython frame)
                         Call func1
    Uninstall tracker
    Install new tracker
                         Perform an allocation
                         Return from func1
                         Return from function_caller
                         Perform an allocation
                         Return from thread_body
    Join thread
    Uninstall tracker
    """
    # GIVEN
    ready_to_replace_tracker = threading.Event()
    tracker_replaced = threading.Event()

    allocator = MemoryAllocator()
    output1 = tmp_path / "test.bin.1"
    output2 = tmp_path / "test.bin.2"

    def thread_body():
        function_caller(func1)
        allocator.valloc(1234)
        allocator.free()

    def func1():
        ready_to_replace_tracker.set()
        tracker_replaced.wait()
        allocator.valloc(1234)
        allocator.free()

    # WHEN
    with Tracker(output1):
        thread = threading.Thread(target=thread_body)
        thread.start()
        ready_to_replace_tracker.wait()

    with Tracker(output2):
        tracker_replaced.set()
        thread.join()

    # THEN
    allocations = list(FileReader(output2).get_allocation_records())

    vallocs = [
        event
        for event in allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    assert len(vallocs) == 2

    first, second = vallocs
    alloc1_funcs = [frame[0] for frame in first.stack_trace()]
    alloc2_funcs = [frame[0] for frame in second.stack_trace()]

    # Cython frames called before tracking started aren't in the Python stack.
    assert alloc1_funcs[:3] == ["valloc", "func1", "thread_body"]

    # But we do get pop events for them if tracking was enabled when they were
    # pushed. Ensure we handle these unexpected pops.
    assert alloc2_funcs[:2] == ["valloc", "thread_body"]


def test_allocation_after_unsetting_profile_function(tmp_path):
    """After tracking starts, unset the profile function then allocate.

    Make sure that the stack for the allocation is unknown.
    """
    # GIVEN
    allocator = MemoryAllocator()
    output = tmp_path / "test.bin"

    def func():
        allocator.valloc(1234)
        allocator.free()
        sys.setprofile(None)
        allocator.valloc(1234)
        allocator.free()

    # WHEN
    with Tracker(output):
        func()

    # THEN
    allocations = list(FileReader(output).get_allocation_records())

    vallocs = [
        event
        for event in allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    assert len(vallocs) == 2

    first, second = vallocs
    alloc1_funcs = [frame[0] for frame in first.stack_trace()]
    alloc2_funcs = [frame[0] for frame in second.stack_trace()]

    assert alloc1_funcs == [
        "valloc",
        "func",
        "test_allocation_after_unsetting_profile_function",
    ]
    assert alloc2_funcs == []


def test_allocation_in_thread_after_unsetting_profile_function(tmp_path):
    """In a thread, unset the profile function then allocate.

    Make sure that the stack for the allocation is unknown.
    """
    # GIVEN
    allocator = MemoryAllocator()
    output = tmp_path / "test.bin"

    def func():
        allocator.valloc(1234)
        allocator.free()
        sys.setprofile(None)
        allocator.valloc(1234)
        allocator.free()

    # WHEN
    with Tracker(output):
        thread = threading.Thread(target=func)
        thread.start()
        thread.join()

    # THEN
    allocations = list(FileReader(output).get_allocation_records())

    vallocs = [
        event
        for event in allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    assert len(vallocs) == 2

    first, second = vallocs
    alloc1_funcs = [frame[0] for frame in first.stack_trace()]
    alloc2_funcs = [frame[0] for frame in second.stack_trace()]

    assert alloc1_funcs[:2] == ["valloc", "func"]
    assert alloc2_funcs == []


class TestMmap:
    @classmethod
    def allocating_function(cls):
        with mmap.mmap(-1, length=2048, access=mmap.ACCESS_WRITE) as mmap_obj:
            mmap_obj[0:100] = b"a" * 100

    @pytest.mark.valgrind
    def test_mmap(self, tmpdir):
        # GIVEN / WHEN
        output = Path(tmpdir) / "test.bin"
        with Tracker(output):
            TestMmap.allocating_function()

        # THEN
        records = list(FileReader(output).get_allocation_records())

        assert len(records) >= 2

        mmap_record = next(
            (record for record in records if AllocatorType.MMAP == record.allocator),
            None,
        )
        assert mmap_record is not None
        assert "allocating_function" in {
            element[0] for element in mmap_record.stack_trace()
        }

        munmap_record = next(
            (record for record in records if AllocatorType.MUNMAP == record.allocator),
            None,
        )
        assert munmap_record is not None
        with pytest.raises(NotImplementedError):
            munmap_record.stack_trace()

    @pytest.mark.valgrind
    def test_mmap_in_thread(self, tmpdir):
        # GIVEN / WHEN
        output = Path(tmpdir) / "test.bin"

        def custom_trace_fn():  # pragma: no cover
            pass

        try:
            threading.setprofile(custom_trace_fn)
            t = threading.Thread(target=TestMmap.allocating_function)
            with Tracker(output):
                t.start()
                t.join()
        finally:
            profile_hook = threading._profile_hook
            threading.setprofile(None)

        # THEN
        assert threading._profile_hook is None
        assert profile_hook == custom_trace_fn
        records = list(FileReader(output).get_allocation_records())

        assert len(records) >= 2

        mmap_record = next(
            (
                record
                for record in records
                if AllocatorType.MMAP == record.allocator and record.size == 2048
            ),
            None,
        )
        assert mmap_record is not None
        assert "allocating_function" in {
            element[0] for element in mmap_record.stack_trace()
        }

        munmap_record = next(
            (record for record in records if AllocatorType.MUNMAP == record.allocator),
            None,
        )
        assert munmap_record is not None
        with pytest.raises(NotImplementedError):
            munmap_record.stack_trace()



import operator
from functools import reduce
from itertools import chain


def fibonacci(length):
    # edge cases
    if length < 1:
        return
    if length == 1:
        yield 1
        return

    left = right = 1
    yield left
    yield right

    for _ in range(length - 2):
        left, right = right, left + right
        yield right


def generate_fibonacci_hash(length_1, length_2, length_3):
    # We could have used sum(...) here instead of reduce(operator.add, ...),
    # but we choose to use reduce since it yields a more descriptive example
    # of the generated flamegraph for this specific example
    return (
        reduce(
            operator.add,
            chain(fibonacci(length_1), fibonacci(length_2), fibonacci(length_3)),
            0,
        )
        % 10000
    )


if __name__ == "__main__":
    # DO NOT CHANGE
    LENGTH_OF_SEQUENCE_1 = 33333
    LENGTH_OF_SEQUENCE_2 = 30000
    LENGTH_OF_SEQUENCE_3 = 34567
    # DO NOT CHANGE
    generate_fibonacci_hash(
        LENGTH_OF_SEQUENCE_1, LENGTH_OF_SEQUENCE_2, LENGTH_OF_SEQUENCE_3
    )



from _memray.records cimport AllocationRecord
from _memray.records cimport CodeObjectInfo
from _memray.records cimport FileFormat
from _memray.records cimport FramePop
from _memray.records cimport FramePush
from _memray.records cimport HeaderRecord
from _memray.records cimport ImageSegments
from _memray.records cimport MemoryRecord
from _memray.records cimport ObjectRecord
from _memray.records cimport ThreadRecord
from _memray.records cimport UnresolvedNativeFrame
from _memray.records cimport code_object_id_t
from _memray.records cimport thread_id_t
from _memray.sink cimport Sink
from libcpp cimport bool
from libcpp.memory cimport unique_ptr
from libcpp.string cimport string
from libcpp.utility cimport pair
from libcpp.vector cimport vector


cdef extern from *:
    ctypedef struct PyCodeObject:
        pass

    ctypedef struct PyFrameObject:
        pass


cdef extern from "compat.h" namespace "memray::compat":
    char* codeGetLinetable(PyCodeObject*, size_t*)
    int frameGetLasti(PyFrameObject*)


cdef extern from "record_writer.h" namespace "memray::tracking_api":
    cdef cppclass RecordWriter:
        bool writeRecord(const MemoryRecord& record) except+
        bool writeRecord(const pair[code_object_id_t, CodeObjectInfo]& record) except+
        bool writeRecord(const UnresolvedNativeFrame& record) except+
        bool writeMappings(const vector[ImageSegments]& mappings) except+
        bool writeThreadSpecificRecord(thread_id_t tid, const FramePop& record) except+
        bool writeThreadSpecificRecord(thread_id_t tid, const FramePush& record) except+
        bool writeThreadSpecificRecord(thread_id_t tid, const AllocationRecord& record) except+
        bool writeThreadSpecificRecord(thread_id_t tid, const ThreadRecord& record) except+
        bool writeThreadSpecificRecord(thread_id_t tid, const ObjectRecord& record) except+
        bool writeHeader(bool seek_to_start) except+
        bool writeTrailer() except+
        void setMainTidAndSkippedFrames(thread_id_t main_tid, size_t skipped_frames_on_main_tid) except+
        unique_ptr[RecordWriter] cloneInChildProcess() except+

    cdef unique_ptr[RecordWriter] createRecordWriter(
        unique_ptr[Sink],
        string command_line,
        bool native_trace,
        FileFormat file_format,
        bool trace_python_allocators,
        bool track_object_lifetimes,
        string libdest,
        vector[string] site_packages,
        vector[string] sys_path,
    ) except+



import subprocess
import sys
import textwrap
from pathlib import Path

import pytest

from memray import AllocatorType
from memray import FileReader
from tests.utils import filter_relevant_allocations

pytestmark = pytest.mark.skipif(
    sys.version_info >= (3, 14), reason="Greenlet does not yet support Python 3.14"
)


def test_integration_with_greenlet(tmpdir):
    """Verify that we can track Python stacks when greenlet is in use."""
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    subprocess_code = textwrap.dedent(
        f"""
        import greenlet

        from memray import Tracker
        from memray._test import MemoryAllocator


        def apple():
            banana()


        def banana():
            allocator.valloc(1024 * 10)
            animal.switch()
            allocator.valloc(1024 * 30)


        def ant():
            allocator.valloc(1024 * 20)
            fruit.switch()
            allocator.valloc(1024 * 40)
            bat()
            allocator.valloc(1024 * 60)


        def bat():
            allocator.valloc(1024 * 50)


        def test():
            fruit.switch()
            assert fruit.dead
            animal.switch()
            assert animal.dead
            allocator.valloc(1024 * 70)


        allocator = MemoryAllocator()
        output = "{output}"

        with Tracker(output):
            fruit = greenlet.greenlet(apple)
            animal = greenlet.greenlet(ant)
            test()
        """
    )

    # WHEN
    subprocess.run([sys.executable, "-Xdev", "-c", subprocess_code], timeout=5)

    # THEN
    reader = FileReader(output)
    records = list(reader.get_allocation_records())
    vallocs = [
        record
        for record in filter_relevant_allocations(records)
        if record.allocator == AllocatorType.VALLOC
    ]

    def stack(alloc):
        return [frame[0] for frame in alloc.stack_trace()]

    assert stack(vallocs[0]) == ["valloc", "banana", "apple"]
    assert vallocs[0].size == 10 * 1024

    assert stack(vallocs[1]) == ["valloc", "ant"]
    assert vallocs[1].size == 20 * 1024

    assert stack(vallocs[2]) == ["valloc", "banana", "apple"]
    assert vallocs[2].size == 30 * 1024

    assert stack(vallocs[3]) == ["valloc", "ant"]
    assert vallocs[3].size == 40 * 1024

    assert stack(vallocs[4]) == ["valloc", "bat", "ant"]
    assert vallocs[4].size == 50 * 1024

    assert stack(vallocs[5]) == ["valloc", "ant"]
    assert vallocs[5].size == 60 * 1024

    assert stack(vallocs[6]) == ["valloc", "test", "<module>"]
    assert vallocs[6].size == 70 * 1024


def test_importing_greenlet_after_tracking_starts(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    subprocess_code = textwrap.dedent(
        f"""
        from memray import Tracker
        from memray._test import MemoryAllocator


        def apple():
            banana()


        def banana():
            allocator.valloc(1024 * 10)
            animal.switch()
            allocator.valloc(1024 * 30)


        def ant():
            allocator.valloc(1024 * 20)
            fruit.switch()
            allocator.valloc(1024 * 40)
            bat()
            allocator.valloc(1024 * 60)


        def bat():
            allocator.valloc(1024 * 50)


        def test():
            fruit.switch()
            assert fruit.dead
            animal.switch()
            assert animal.dead
            allocator.valloc(1024 * 70)


        allocator = MemoryAllocator()
        output = "{output}"

        with Tracker(output):
            import greenlet

            fruit = greenlet.greenlet(apple)
            animal = greenlet.greenlet(ant)
            test()
        """
    )

    # WHEN
    subprocess.run([sys.executable, "-Xdev", "-c", subprocess_code], timeout=5)

    # THEN
    reader = FileReader(output)
    records = list(reader.get_allocation_records())
    vallocs = [
        record
        for record in filter_relevant_allocations(records)
        if record.allocator == AllocatorType.VALLOC
    ]

    def stack(alloc):
        return [frame[0] for frame in alloc.stack_trace()]

    assert stack(vallocs[0]) == ["valloc", "banana", "apple"]
    assert vallocs[0].size == 10 * 1024

    assert stack(vallocs[1]) == ["valloc", "ant"]
    assert vallocs[1].size == 20 * 1024

    assert stack(vallocs[2]) == ["valloc", "banana", "apple"]
    assert vallocs[2].size == 30 * 1024

    assert stack(vallocs[3]) == ["valloc", "ant"]
    assert vallocs[3].size == 40 * 1024

    assert stack(vallocs[4]) == ["valloc", "bat", "ant"]
    assert vallocs[4].size == 50 * 1024

    assert stack(vallocs[5]) == ["valloc", "ant"]
    assert vallocs[5].size == 60 * 1024

    assert stack(vallocs[6]) == ["valloc", "test", "<module>"]
    assert vallocs[6].size == 70 * 1024

    # 0 and 2 are fruit, 1 and 3 and 4 and 5 are animal, 6 is main.
    assert vallocs[0].tid != vallocs[1].tid != vallocs[6].tid
    assert vallocs[0].tid == vallocs[2].tid
    assert vallocs[1].tid == vallocs[3].tid == vallocs[4].tid == vallocs[5].tid


def test_uninstall_profile_in_greenlet(tmpdir):
    """Verify that memray handles profile function changes in greenlets correctly."""
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    subprocess_code = textwrap.dedent(
        f"""
        import greenlet
        import sys

        from memray import Tracker
        from memray._test import MemoryAllocator

        def foo():
            bar()
            allocator.valloc(1024 * 10)

        def bar():
            baz()

        def baz():
            sys.setprofile(None)
            other.switch()

        def test():
            allocator.valloc(1024 * 70)
            main_greenlet.switch()


        allocator = MemoryAllocator()
        output = "{output}"

        with Tracker(output):
            main_greenlet = greenlet.getcurrent()
            other = greenlet.greenlet(test)
            foo()

        """
    )

    # WHEN
    subprocess.run([sys.executable, "-Xdev", "-c", subprocess_code], timeout=5)

    # THEN
    reader = FileReader(output)
    records = list(reader.get_allocation_records())
    vallocs = [
        record
        for record in filter_relevant_allocations(records)
        if record.allocator == AllocatorType.VALLOC
    ]

    def stack(alloc):
        return [frame[0] for frame in alloc.stack_trace()]

    # Verify allocations and their stack traces (which should be empty
    # because we remove the tracking function)
    assert len(vallocs) == 2

    assert stack(vallocs[0]) == []
    assert vallocs[0].size == 70 * 1024

    assert stack(vallocs[1]) == []
    assert vallocs[1].size == 10 * 1024

    # Verify thread IDs
    main_tid = vallocs[0].tid  # inner greenlet
    outer_tid = vallocs[1].tid  # outer greenlet
    assert main_tid == outer_tid



import os
import sysconfig
from distutils.core import Extension
from distutils.core import setup

ROOT = os.path.realpath(os.path.dirname(__file__))
LDSHARED = os.environ.get("LDSHARED", sysconfig.get_config_var("LDSHARED"))
if LDSHARED:
    LDSHARED = LDSHARED.replace("--strip-all", "-g")
    os.environ["LDSHARED"] = LDSHARED

setup(
    name="misbehaving",
    version="0.0",
    ext_modules=[
        Extension(
            "misbehaving",
            language="c++",
            sources=[os.path.join(ROOT, "misbehaving.cpp")],
        ),
    ],
    zip_safe=False,
)



"""Test the performance of pprint.PrettyPrinter.

This benchmark was available as `python -m pprint` until Python 3.12.

Authors: Fred Drake (original), Oleg Iarygin (pyperformance port).
"""

from pprint import PrettyPrinter

printable = [("string", (1, 2), [3, 4], {5: 6, 7: 8})] * 100_000
p = PrettyPrinter()


def run_benchmark():
    if hasattr(p, "_safe_repr"):
        p._safe_repr(printable, {}, None, 0)
    p.pformat(printable)


if __name__ == "__main__":
    run_benchmark()



import json
import sys

import pyperf
from memray_helper import get_tracker


EMPTY = ({}, 2000)
SIMPLE_DATA = {
    "key1": 0,
    "key2": True,
    "key3": "value",
    "key4": "foo",
    "key5": "string",
}
SIMPLE = (SIMPLE_DATA, 1000)
NESTED_DATA = {
    "key1": 0,
    "key2": SIMPLE[0],
    "key3": "value",
    "key4": SIMPLE[0],
    "key5": SIMPLE[0],
    "key": "\u0105\u0107\u017c",
}
NESTED = (NESTED_DATA, 1000)
HUGE = ([NESTED[0]] * 1000, 1)

CASES = ["EMPTY", "SIMPLE", "NESTED", "HUGE"]


def bench_json_dumps(data):
    with get_tracker():
        for obj, count_it in data:
            for _ in count_it:
                json.dumps(obj)


def add_cmdline_args(cmd, args):
    if args.cases:
        cmd.extend(("--cases", args.cases))


def main():
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.argparser.add_argument(
        "--cases",
        help="Comma separated list of cases. Available cases: %s. By default, run all cases."
        % ", ".join(CASES),
    )
    runner.metadata["description"] = "Benchmark json.dumps()"

    args = runner.parse_args()
    if args.cases:
        cases = []
        for case in args.cases.split(","):
            case = case.strip()
            if case:
                cases.append(case)
        if not cases:
            print("ERROR: empty list of cases")
            sys.exit(1)
    else:
        cases = CASES

    data = []
    for case in cases:
        obj, count = globals()[case]
        data.append((obj, range(count)))

    runner.bench_func("json_dumps", bench_json_dumps, data)


if __name__ == "__main__":
    main()



from datetime import datetime

import pytest

from memray import Metadata
from memray._memray import FileFormat
from memray.reporters.templates import get_report_title
from memray.reporters.templates import render_report


@pytest.mark.parametrize(
    ["kind", "show_memory_leaks", "inverted", "expected"],
    [
        ("flamegraph", False, False, "flamegraph report"),
        ("flamegraph", True, False, "flamegraph report (memory leaks)"),
        ("table", False, False, "table report"),
        ("table", True, False, "table report (memory leaks)"),
        ("flamegraph", False, True, "inverted flamegraph report"),
        ("flamegraph", True, True, "inverted flamegraph report (memory leaks)"),
    ],
)
def test_title_for_regular_report(kind, show_memory_leaks, inverted, expected):
    assert (
        get_report_title(
            kind=kind, show_memory_leaks=show_memory_leaks, inverted=inverted
        )
        == expected
    )


@pytest.mark.parametrize(
    "kind",
    ["flamegraph", "table"],
)
def test_html_report_escaping(kind):
    """Test that command line arguments are properly escaped."""
    # GIVEN
    metadata = Metadata(
        start_time=datetime(2024, 1, 1, 0, 0, 0),
        end_time=datetime(2024, 1, 1, 0, 1, 0),
        total_allocations=100,
        total_frames=10,
        peak_memory=1024,
        command_line="python test.py </code>",
        pid=12345,
        main_thread_id=1,
        python_allocator="pymalloc",
        has_native_traces=False,
        trace_python_allocators=False,
        file_format=FileFormat.ALL_ALLOCATIONS,
    )

    # WHEN
    html_output = render_report(
        kind=kind,
        data=[],
        metadata=metadata,
        memory_records=[],
        show_memory_leaks=False,
        merge_threads=False,
        inverted=False,
        no_web=True,
    )

    # THEN
    assert html_output.count("<code>") == html_output.count("</code>")
    assert "python test.py &lt;/code&gt;" in html_output
